;;; accent-voices.el --- Define various device independent voices in terms of Accent codes.
;;; $Id$
;;; Author: garym@linux.ca
;;; Description:  Module to set up Accent voices and personalities
;;; Keywords: Voice, Personality, Accent
;;{{{  LCD Archive entry:

;;; LCD Archive Entry:
;;; emacspeak| T. V. Raman |raman@cs.cornell.edu
;;; AccentSA speech interface to Emacs |
;;; $Date$ |
;;;  $Revision$ |
;;; Location undetermined
;;;

;;}}}
;;{{{  Copyright:
;;;Copyright (C) 2005 Gary Murphy
;;; All Rights Reserved.
;;;
;;; This file is not part of GNU Emacs, but the same permissions apply.
;;;
;;; GNU Emacs is free software; you can redistribute it and/or modify
;;; it under the terms of the GNU General Public License as published by
;;; the Free Software Foundation; either version 2, or (at your option)
;;; any later version.
;;;
;;; GNU Emacs is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;; GNU General Public License for more details.
;;;
;;; You should have received a copy of the GNU General Public License
;;; along with GNU Emacs; see the file COPYING.  If not, write to
;;; the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

;;}}}
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;{{{  Introduction:

;;; Commentary:
;;; This module defines the various voices used in voice-lock mode.
;;; This module is AccentSA specific.

;;}}}
;;{{{ required modules

;;; Code:
(require 'cl)
(declaim  (optimize  (safety 0) (speed 3)))

;; Implements the CSS 'Speech Stylesheets'
(require 'acss-structure)
;;}}}
;;{{{  voice table

(defcustom accent-default-speech-rate 60
  "Default speech rate for AccentSA."
  :group 'tts
  :type 'integer)


(defvar accent-default-voice-string "\eP5\eV5"
  "Accent string for  default voice --set to be a no-op.")

(defvar accent-voice-table (make-hash-table)
  "Association between symbols and strings to set Accent voices.
The string can set any Accent parameter.")

(defsubst accent-define-voice (name command-string)
  "Define a Accent voice named NAME.
This voice will be set   by sending the string
COMMAND-STRING to the Accent synth."
  (declare (special accent-voice-table ))
  (puthash  name command-string  accent-voice-table))

;; dectalk has no *-get-voice-command-internal but outloud does
;; and uses it in the get-voice-command before setting from dtk-speech-rate.
;; is it needed?

(defsubst accent-get-voice-command (name)
  "Retrieve command string for  voice NAME."
  (declare (special accent-voice-table ))
  (cond
   ((listp name)
    (mapconcat #'accent-get-voice-command name " "))
   (t (or  (gethash name accent-voice-table)
           accent-default-voice-string))))

(defsubst accent-voice-defined-p (name)
  "Check if there is a voice named NAME defined."
  (declare (special accent-voice-table ))
  (gethash name accent-voice-table ))

;;}}}
;;{{{ voice definitions

;;; the nine predefined voices:
(accent-define-voice 'paul   "\eP5\eV5 ")
(accent-define-voice 'harry  "\eP4\eV5 ")
(accent-define-voice 'dennis "\eP4\eV4 ")
(accent-define-voice 'frank  "\eP3\eV4 ")
(accent-define-voice 'betty  "\eP5\eV6 ")
(accent-define-voice 'ursula "\eP6\eV6 ")
(accent-define-voice 'rita   "\eP6\eV7 ")
(accent-define-voice 'wendy  "\eP7\eV7 ")
(accent-define-voice 'kit    "\eP8\eV8 ")

;;}}}
;;{{{  the inaudible voice
;;; no special code needed --handled by Emacspeak engine.

(accent-define-voice 'inaudible "\eP0\eV0 ")

;;}}}
;;{{{  Mapping css parameters to Accent codes

;;{{{ voice family codes

(defvar accent-family-table nil
  "Association list of Accent voice names and control codes.")

(defsubst accent-set-family-code (name code)
  "Set control code for voice family NAME  to CODE."
  (declare (special accent-family-table))
  (when (stringp name)
    (setq name (intern name)))
  (setq accent-family-table
        (cons (list name code )
              accent-family-table)))

(defsubst accent-get-family-code (name)
  "Get control code for voice family NAME."
  (declare (special accent-family-table ))
  (when (stringp name)
    (setq name (intern name )))
  (or (cadr (assq  name accent-family-table))
      ""))

(accent-set-family-code 'paul ":np")
(accent-set-family-code 'harry ":nh")
(accent-set-family-code 'dennis ":nd")
(accent-set-family-code 'frank ":nf")
(accent-set-family-code 'betty ":nb")
(accent-set-family-code 'ursula ":nu")
(accent-set-family-code 'wendy ":nw")
(accent-set-family-code 'rita ":nr")
(accent-set-family-code 'kid ":nk")

;;}}}
;;{{{  hash table for mapping families to their dimensions

(defvar accent-css-code-tables (make-hash-table)
  "Hash table holding vectors of Accent codes.
Keys are symbols of the form <FamilyName-Dimension>.
Values are vectors holding the control codes for the 10 settings.")

(defsubst accent-css-set-code-table (family dimension table)
  "Set up voice FAMILY.
Argument DIMENSION is the dimension being set,
and TABLE gives the values along that dimension."
  (declare (special accent-css-code-tables))
  (let ((key (intern (format "%s-%s" family dimension))))
    (puthash  key table accent-css-code-tables )))

(defsubst accent-css-get-code-table (family dimension)
  "Retrieve table of values for specified FAMILY and DIMENSION."
  (declare (special accent-css-code-tables))
  (let ((key (intern (format "%s-%s" family dimension))))
    (gethash key accent-css-code-tables)))

;;}}}
;;{{{ volume

;;; Note:volume settings not implemented for Accents.
(defvar accent-gain-table (make-vector  10 "")
  "Maps CSS volume settings to actual synthesizer codes.")

;;}}}
;;{{{  average pitch

;;; Average pitch for standard male voice is 122hz --this is mapped to
;;; a setting of 5.
;;; Average pitch varies inversely with speaker head size --a child
;;; has a small head and a higher pitched voice.
;;; We change parameter head-size in conjunction with average pitch to
;;; produce a more natural change on the Accent.

;;{{{  paul average pitch

(let ((table (make-vector 10 "")))
  (mapcar
   (function
    (lambda (setting)
      (aset table
	    (first setting)
	    (format "\eP%s\eV%s "
		    (second setting)
		    (third setting)))))
   '(
     (0 3 2)
     (1 3 3)
     (2 4 3)
     (3 4 4)
     (4 5 4)
     (5 5 5)
     (6 5 6)
     (7 6 6)
     (8 6 7)
     (9 7 7)
     ))
  (accent-css-set-code-table 'paul 'average-pitch table ))

;;}}}
;;{{{  harry average pitch
;;; Harry  has a big head --and a lower pitch for the middle setting 

(let ((table (make-vector 10 "")))
  (mapcar
   (function
    (lambda (setting)
      (aset table
	    (first setting)
	    (format "\eP%s\eV%s "
		    (second setting)
		    (third setting)))))
   '(
     (0 2 1)
     (1 2 2)
     (2 3 2)
     (3 3 3)
     (4 4 3)
     (5 4 4)
     (6 4 5)
     (7 5 5)
     (8 5 6)
     (9 6 6)
     ))
  (accent-css-set-code-table 'harry 'average-pitch table ))

;;}}}
;;{{{  betty average pitch

(let ((table (make-vector 10 "")))
  (mapcar
   (function
    (lambda (setting)
      (aset table
	    (first setting)
	    (format "\eP%s\eV%s "
		    (second setting)
		    (third setting)))))
   '(
     (0 4 2)
     (1 4 3)
     (2 5 3)
     (3 5 4)
     (4 6 4)
     (5 6 5)
     (6 6 6)
     (7 7 6)
     (8 7 7)
     (9 8 7)
     ))
  (accent-css-set-code-table 'betty 'average-pitch table ))

;;}}}

(defsubst accent-get-average-pitch-code (value family)
  "Get  AVERAGE-PITCH for specified VALUE and  FAMILY."
  (or family (setq family 'paul))
  (if value 
      (aref (accent-css-get-code-table family 'average-pitch)
	    value)
    ""))

;;}}}
;;{{{  pitch range

;;;  Standard pitch range is 100 and is  mapped to
;;; a setting of 5.
;;; A value of 0 produces a flat monotone voice --maximum value of 250
;;; produces a highly animated voice.  With the Accent, 0 is animated,
;;; 1 is monotone, 2-4 are the range between them.
;;; Additionally, the tabpe sets the assertiveness of the voice so the
;;; voice is less assertive at lower pitch ranges, but this is not
;;; supported by the AccentSA.
;;{{{  paul pitch range

(let ((table (make-vector 10 "")))
  (mapcar
   (function
    (lambda (setting)
      (aset table
	    (first setting)
	    (format "\eM%s "
		    (second setting)
		    (third setting))))) ;; not implemented in AccentSA
   '(
     (0 1 0)
     (1 4 10)
     (2 4 20)
     (3 4 30)
     (4 3 40 )
     (5 3 50 )
     (6 3 0)
     (7 2 70)
     (8 2 80)
     (9 0 100)
     ))
  (accent-css-set-code-table 'paul 'pitch-range table ))

;;}}}
;;{{{  harry pitch range

(let ((table (make-vector 10 "")))
  (mapcar
   (function
    (lambda (setting)
      (aset table
	    (first setting)
	    (format "\eM%s "
		    (second setting)
		    (third setting)))))
   '(
     (0 1 0)
     (1 4 10)
     (2 4 20)
     (3 4 30)
     (4 3 40 )
     (5 3 50 )
     (6 3 0)
     (7 2 70)
     (8 2 80)
     (9 0 100)
     ))
  (accent-css-set-code-table 'harry 'pitch-range table ))

;;}}}
;;{{{  betty pitch range

(let ((table (make-vector 10 "")))
  (mapcar
   (function
    (lambda (setting)
      (aset table
	    (first setting)
	    (format "\eM%s "
		    (second setting)
		    (third setting)))))
   '(
     (0 1 0)
     (1 4 10)
     (2 4 20)
     (3 4 30)
     (4 3 40 )
     (5 3 50 )
     (6 3 0)
     (7 2 70)
     (8 2 80)
     (9 0 100)
     ))
  (accent-css-set-code-table 'betty 'pitch-range table ))

;;}}}
(defsubst accent-get-pitch-range-code (value family)
  "Get pitch-range code for specified VALUE and FAMILY."
  (or family (setq family 'paul))
  (if value 
      (aref (accent-css-get-code-table family 'pitch-range)
	    value)
    ""))

;;}}}
;;{{{  stress

;;; On the DTK we can vary four parameters:
;;; The hat rise which controls the overall shape of the F0 contour
;;; for sentence level intonation and stress,
;;; The stress rise that controls the level of stress on stressed
;;; syllables,
;;; the baseline fall for paragraph level intonation
;;; and the quickness --a parameter that controls whether the final
;;; frequency targets are completely achieved in the phonetic
;;; transitions.
;;; The Accent has none of these controls.
;;{{{  paul stress

(let ((table (make-vector 10 "")))
  (mapcar
   (function
    (lambda (setting)
      (aset table
	    (first setting)
	    (format ""
		    (second setting)
		    (third setting)
		    (fourth setting)
		    (fifth setting)))))
   '(
     (0  0 0 0 0)
     (1 3 6  20 3)
     (2 6 12  40 6)
     (3 9 18  60 9 )
     (4 12 24 80 14)
     (5 18 32  100 18)
     (6 34 50 100 20)
     (7 48  65 100 35)
     (8 63 82 100 60)
     (9 80  90 100  40)
     ))
  (accent-css-set-code-table 'paul 'stress table))

;;}}}
;;{{{  harry stress

(let ((table (make-vector 10 "")))
  (mapcar
   (function
    (lambda (setting)
      (aset table
	    (first setting)
	    (format ""
		    (second setting)
		    (third setting)
		    (fourth setting)
		    (fifth setting)))))
   '(
     (0  0 0 0 0)
     (1 4 6 2 2 )
     (2 8 12 4 4 )
     (3 12 18 6 6 )
     (4 16 24 8 8 )
     (5 20 30 10 9)
     (6 40  48 32 16)
     (7 60 66 54 22)
     (8 80 78 77 34)
     (9 100 100 100 40)
     ))
  (accent-css-set-code-table 'harry 'stress table))

;;}}}
;;{{{  betty stress

(let ((table (make-vector 10 "")))
  (mapcar
   (function
    (lambda (setting)
      (aset table
	    (first setting)
	    (format ""
		    (second setting)
		    (third setting)
		    (fourth setting)
		    (fifth setting)))))
   '(
     (0  1 1 0 0)
     (1 3 4 11 0)
     (2 5 8 22 0)
     (3 8 12 33 0 )
     (4 11  16 44 0)
     (5 14 20 55 0)
     (6 35 40 65 10)
     (7 56 80 75 20)
     (8 77 90 85 30)
     (9 100 100 100 40)
     ))
  (accent-css-set-code-table 'betty 'stress table))

;;}}}
(defsubst accent-get-stress-code (value family)
  (or family (setq family 'paul ))
  (if value 
      (aref (accent-css-get-code-table family 'stress)
	    value)
    ""))

;;}}}
;;{{{  richness

;;; Smoothness and richness vary inversely.
;;; a  maximally smooth voice produces a quieter effect
;;; a rich voice is "bright" in contrast.
;;; The Accent has none of these controls.
;;{{{  paul richness

(let ((table (make-vector 10 "")))
  (mapcar
   (function
    (lambda (setting)
      (aset table (first setting)
	    (format ""
		    (second setting)
		    (third setting)))))
   '(
     (0 0 100)
     (1 14 80)
     (2 28 60)
     (3 42 40 )
     (4 56 20)
     (5 70  3 )
     (6 60 24 )
     (7 70 16)
     (8 80 8 20)
     (9 100  0)
     ))
  (accent-css-set-code-table 'paul 'richness table))

;;}}}
;;{{{  harry richness

(let ((table (make-vector 10 "")))
  (mapcar
   (function
    (lambda (setting)
      (aset table (first setting)
	    (format ""
		    (second setting)
		    (third setting)))))
   '(
     (0 100 0)
     (1 96 3)
     (2 93 6)
     (3 90 9)
     (4 88 11)
     (5 86 12)
     (6 60 24 )
     (7 40 44)
     (8 20 65)
     (9 0 70)
     ))
  (accent-css-set-code-table 'harry 'richness table))

;;}}}
;;{{{  betty richness

(let ((table (make-vector 10 "")))
  (mapcar
   (function
    (lambda (setting)
      (aset table (first setting)
	    (format ""
		    (second setting)
		    (third setting)))))
   '(
     (0 0 100)
     (1 8 76)
     (2 16 52)
     (3 24  28)
     (4 32 10)
     (5 40 4)
     (6 50 3 )
     (7 65 3)
     (8 80 8 2)
     (9 100  0)
     ))
  (accent-css-set-code-table 'betty 'richness table))

;;}}}

(defsubst accent-get-richness-code (value family)
  (or family (setq family 'paul))
  (if value 
      (aref (accent-css-get-code-table family 'richness)
	    value)
    ""))

;;}}}
;;{{{  punctuations
;;; Accent has only full punctuation or normal text mode.

(defsubst accent-get-punctuations-code (value)
  "Return string needed to set specified punctuations mode."
  (if value 
      "\eOP\eOR"
    "\eOp"))

;;}}}

;;}}}
;;{{{  accent-define-voice-from-speech-style

(defun accent-define-voice-from-speech-style (name style)
  "Define NAME to be a Accent voice as specified by settings in STYLE."
  (let* ((family(acss-family style))
	 (command
	  (concat
		  (accent-get-voice-command family)
                  (accent-get-punctuations-code (acss-punctuations style))
                  (when (or (acss-average-pitch style)
                            (acss-pitch-range style)
                            (acss-stress style )
                            (acss-richness style))
		    (concat 
			    (accent-get-average-pitch-code (acss-average-pitch style) family)
			    (accent-get-pitch-range-code (acss-pitch-range style) family)
			    (accent-get-stress-code (acss-stress style ) family)
			    (accent-get-richness-code (acss-richness style) family)))
		 )))
    (accent-define-voice name command)))

;;}}}
;;{{{ list voices 

(defun accent-list-voices ()
  "List defined voices."
  (declare (special accent-voice-table))
  (loop for k being the hash-keys of accent-voice-table 
	collect   k))

;;}}}
;;{{{ configurater 

(defun accent-configure-tts ()
  "Configures TTS environment to use Accent family of synthesizers."
  (declare (special  accent-default-speech-rate
                     tts-default-speech-rate))
  (fset 'tts-list-voices 'accent-list-voices)
  (fset 'tts-voice-defined-p 'accent-voice-defined-p)
  (fset 'tts-get-voice-command 'accent-get-voice-command)
  (fset 'tts-voice-defined-p 'accent-voice-defined-p)
  (fset 'tts-define-voice-from-speech-style 'accent-define-voice-from-speech-style)
  (setq tts-default-speech-rate accent-default-speech-rate)
  (set-default 'tts-default-speech-rate accent-default-speech-rate))

;;}}}
(provide 'accent-voices)
;;{{{  emacs local variables

;;; local variables:
;;; folded-file: t
;;; byte-compile-dynamic: t
;;; end:

;;}}}
